@@ -12,21 +12,9 @@ except ImportError:
     have_sqlalchemy = False
 
 
-def _postgresql_mean_expr(dbmodel, expression):
-    return (
-        "AVG(" + dbmodel.expr_to_sql(expression.args[0], want_inline_parens=False) + ")"
-    )
-
-
-def _postgresql_size_expr(dbmodel, expression):
-    return "SUM(1)"
-
-
 # map from op-name to special SQL formatting code
 PostgreSQL_formatters = {
     "___": lambda dbmodel, expression: expression.to_python(),
-    "mean": _postgresql_mean_expr,
-    "size": _postgresql_size_expr,
 }
 
 
@@ -53,7 +41,7 @@ def example_handle():
     if not have_sqlalchemy:
         return None
     db_handle = PostgreSQLModel().db_handle(
-        sqlalchemy.engine.create_engine(r'postgresql://johnmount@localhost/johnmount')
+        sqlalchemy.engine.create_engine(r"postgresql://johnmount@localhost/johnmount")
     )
     db_handle.db_model.prepare_connection(db_handle.conn)
     return db_handle
