@@ -11,11 +11,11 @@ import data_algebra.data_model
 import data_algebra.db_model
 import data_algebra.pandas_model
 import data_algebra.expr_rep
-import data_algebra.user_fn
 from data_algebra.data_ops_types import *
 import data_algebra.data_ops_utils
 import data_algebra.near_sql
 import data_algebra.OrderedSet
+import data_algebra.util
 
 
 _have_black = False
@@ -48,7 +48,6 @@ class ViewRepresentation(OperatorPlatform, ABC):
 
     column_names: List[str]
     column_set: data_algebra.OrderedSet.OrderedSet
-    column_map: collections.OrderedDict
     sources: List[
         "ViewRepresentation"
     ]  # https://www.python.org/dev/peps/pep-0484/#forward-references
@@ -58,8 +57,7 @@ class ViewRepresentation(OperatorPlatform, ABC):
             column_names = [column_names]
         self.column_names = [c for c in column_names]
         for ci in self.column_names:
-            if not isinstance(ci, str):
-                raise ValueError("non-string column name(s)")
+            assert isinstance(ci, str)
         if len(self.column_names) < 1:
             raise ValueError("no column names")
         self.column_set = data_algebra.OrderedSet.OrderedSet()
@@ -71,14 +69,14 @@ class ViewRepresentation(OperatorPlatform, ABC):
             ci: data_algebra.expr_rep.ColumnReference(self, ci)
             for ci in self.column_names
         }
-        self.column_map = collections.OrderedDict(**column_dict)
         if sources is None:
             sources = []
         for si in sources:
-            if not isinstance(si, ViewRepresentation):
-                raise ValueError("all sources must be of class ViewRepresentation")
+            assert isinstance(si, ViewRepresentation)
         self.sources = [si for si in sources]
-        OperatorPlatform.__init__(self, node_name=node_name)
+        OperatorPlatform.__init__(
+            self, node_name=node_name, column_map=collections.OrderedDict(**column_dict)
+        )
 
     def merged_rep_id(self):
         return "node+ " + str(id(self))
@@ -93,10 +91,7 @@ class ViewRepresentation(OperatorPlatform, ABC):
             s = self.sources[i]
             ti = s.get_tables()
             for (k, v) in ti.items():
-                if not isinstance(v, TableDescription):
-                    raise TypeError(
-                        "Expected v to be data_algebra.data_ops.TableDescription"
-                    )
+                assert isinstance(v, TableDescription)
                 if k in tables.keys():
                     if not v.same_table(tables[k]):
                         raise ValueError(
@@ -178,8 +173,10 @@ class ViewRepresentation(OperatorPlatform, ABC):
         self.columns_used()  # for table consistency check/raise
         if pretty:
             strict = True
-        python_str = self.to_python_implementation(
-            indent=indent, strict=strict, print_sources=True
+        python_str = (
+            "(\n"
+            + self.to_python_implementation(indent=indent, strict=strict, print_sources=True)
+            + "\n)\n"
         )
         if pretty:
             python_str = pretty_format_python(python_str, black_mode=black_mode)
@@ -223,31 +220,25 @@ class ViewRepresentation(OperatorPlatform, ABC):
 
     # query generation
 
-    def to_near_sql_implementation(self, db_model, *, using, temp_id_source):
+    def to_near_sql_implementation(self, db_model, *, using, temp_id_source, sql_format_options=None):
         raise NotImplementedError("base method called")
 
     def to_sql(
         self,
         db_model,
         *,
-        pretty=False,
-        annotate=False,
-        encoding=None,
-        sqlparse_options=None,
         temp_tables=None,
-        use_with=False
+        sql_format_options=None,
     ):
         if isinstance(db_model, data_algebra.db_model.DBHandle):
             db_model = db_model.db_model
         assert isinstance(db_model, data_algebra.db_model.DBModel)
+        if sql_format_options is None:
+            sql_format_options = db_model.default_SQL_format_options
         return db_model.to_sql(
             ops=self,
-            pretty=pretty,
-            annotate=annotate,
-            encoding=encoding,
-            sqlparse_options=sqlparse_options,
             temp_tables=temp_tables,
-            use_with=use_with
+            sql_format_options=sql_format_options,
         )
 
     # Pandas realization
@@ -292,16 +283,12 @@ class ViewRepresentation(OperatorPlatform, ABC):
          :return:
          """
 
-        if not isinstance(data_map, dict):
-            raise TypeError("data_map should be a dictionary")
+        assert isinstance(data_map, dict)
         if len(data_map) < 1:
             raise ValueError("Expected data_map to be non-empty")
         if data_model is None:
             data_model = data_algebra.default_data_model
-        if not isinstance(data_model, data_algebra.data_model.DataModel):
-            raise TypeError(
-                "Expected data_model to be derived from data_algebra.data_model.DataModel"
-            )
+        assert isinstance(data_model, data_algebra.data_model.DataModel)
         self.columns_used()  # for table consistency check/raise
         tables = self.get_tables()
         self.check_constraints(
@@ -321,25 +308,18 @@ class ViewRepresentation(OperatorPlatform, ABC):
     def transform(self, X, *, data_model=None, narrow=True):
         if data_model is None:
             data_model = data_algebra.default_data_model
-        if not isinstance(data_model, data_algebra.data_model.DataModel):
-            raise TypeError(
-                "Expected data_model to be derived from data_algebra.data_model.DataModel"
-            )
+        assert isinstance(data_model, data_algebra.data_model.DataModel)
         self.columns_used()  # for table consistency check/raise
         tables = self.get_tables()
         if len(tables) != 1:
             raise ValueError(
-                "transfrom(DataFrame) can only be applied to ops-dags with only one table def"
+                "transform(DataFrame) can only be applied to ops-dags with only one table def"
             )
         k = [k for k in tables.keys()][0]
         # noinspection PyUnresolvedReferences
         if isinstance(X, data_model.pd.DataFrame):
             data_map = {k: X}
-            return self.eval(
-                data_map=data_map,
-                data_model=data_model,
-                narrow=narrow,
-            )
+            return self.eval(data_map=data_map, data_model=data_model, narrow=narrow,)
         raise TypeError("can not apply transform() to type " + str(type(X)))
 
     # composition (used to eliminate intermediate order nodes)
@@ -421,14 +401,10 @@ class ViewRepresentation(OperatorPlatform, ABC):
             reverse=reverse,
         )
 
-    def extend(
-        self, ops, *, partition_by=None, order_by=None, reverse=None, parse_env=None
-    ):
+    def extend(self, ops, *, partition_by=None, order_by=None, reverse=None):
         if (ops is None) or (len(ops) < 1):
             return self
-        parsed_ops = data_algebra.expr_parse.parse_assignments_in_context(
-            ops, self, parse_env=parse_env
-        )
+        parsed_ops = data_algebra.expr_parse.parse_assignments_in_context(ops, self)
         return self.extend_parsed(
             parsed_ops=parsed_ops,
             partition_by=partition_by,
@@ -448,14 +424,12 @@ class ViewRepresentation(OperatorPlatform, ABC):
             return self.sources[0].project_parsed(parsed_ops, group_by=group_by)
         return ProjectNode(source=self, parsed_ops=parsed_ops, group_by=group_by)
 
-    def project(self, ops=None, *, group_by=None, parse_env=None):
+    def project(self, ops=None, *, group_by=None):
         if group_by is None:
             group_by = []
         if ((ops is None) or (len(ops) < 1)) and (len(group_by) < 1):
             raise ValueError("must have ops or group_by")
-        parsed_ops = data_algebra.expr_parse.parse_assignments_in_context(
-            ops, self, parse_env=parse_env
-        )
+        parsed_ops = data_algebra.expr_parse.parse_assignments_in_context(ops, self)
         return self.project_parsed(parsed_ops=parsed_ops, group_by=group_by)
 
     def natural_join(self, b, *, by, jointype):
@@ -466,10 +440,7 @@ class ViewRepresentation(OperatorPlatform, ABC):
         :param jointype: one of 'INNER', 'LEFT', 'RIGHT', 'OUTER', 'FULL', 'CROSS' (case insensitive)
         :return: ops describing join
         """
-        if not isinstance(b, ViewRepresentation):
-            raise TypeError(
-                "expected b to be a data_algebra.dat_ops.ViewRepresentation"
-            )
+        assert isinstance(b, ViewRepresentation)
         if self.is_trivial_when_intermediate():
             return self.sources[0].natural_join(b, by=by, jointype=jointype)
         return NaturalJoinNode(a=self, b=b, by=by, jointype=jointype)
@@ -477,10 +448,7 @@ class ViewRepresentation(OperatorPlatform, ABC):
     def concat_rows(self, b, *, id_column="source_name", a_name="a", b_name="b"):
         if b is None:
             return self
-        if not isinstance(b, ViewRepresentation):
-            raise TypeError(
-                "expected b to be a data_algebra.dat_ops.ViewRepresentation"
-            )
+        assert isinstance(b, ViewRepresentation)
         if self.is_trivial_when_intermediate():
             return self.sources[0].concat_rows(
                 b, id_column=id_column, a_name=a_name, b_name=b_name
@@ -496,14 +464,12 @@ class ViewRepresentation(OperatorPlatform, ABC):
             return self.sources[0].select_rows_parsed(parsed_expr=parsed_expr)
         return SelectRowsNode(source=self, ops=parsed_expr)
 
-    def select_rows(self, expr, *, parse_env=None):
+    def select_rows(self, expr):
         if expr is None:
             return self
         if self.is_trivial_when_intermediate():
-            return self.sources[0].select_rows(expr, parse_env=parse_env)
-        ops = data_algebra.expr_parse.parse_assignments_in_context(
-            {"expr": expr}, self, parse_env=parse_env
-        )
+            return self.sources[0].select_rows(expr)
+        ops = data_algebra.expr_parse.parse_assignments_in_context({"expr": expr}, self)
 
         def r_walk_expr(opv):
             if not isinstance(opv, data_algebra.expr_rep.Expression):
@@ -554,7 +520,9 @@ class ViewRepresentation(OperatorPlatform, ABC):
             return self
         if self.is_trivial_when_intermediate():
             return self.sources[0].convert_records(record_map, temp_namer=temp_namer)
-        return ConvertRecordsNode(source=self, record_map=record_map, temp_namer=temp_namer)
+        return ConvertRecordsNode(
+            source=self, record_map=record_map, temp_namer=temp_namer
+        )
 
 
 # Could also have general query as starting node, but don't see a lot of point to
@@ -566,7 +534,7 @@ class TableDescription(ViewRepresentation):
 
        Example:
            from data_algebra.data_ops import *
-           d = TableDescription('d', ['x', 'y'])
+           d = TableDescription(table_name='d', column_names=['x', 'y'])
            print(d)
     """
 
@@ -577,22 +545,23 @@ class TableDescription(ViewRepresentation):
 
     def __init__(
         self,
+        *,
         table_name,
         column_names,
-        *,
         qualifiers=None,
         sql_meta=None,
         column_types=None,
         head=None,
-        limit_was=None
+        limit_was=None,
+        nrows=None
     ):
         ViewRepresentation.__init__(
             self, column_names=column_names, node_name="TableDescription"
         )
         if table_name is None:
             table_name = ""
-        if (table_name is not None) and (not isinstance(table_name, str)):
-            raise TypeError("table_name must be a string")
+        if table_name is not None:
+            assert isinstance(table_name, str)
         if head is not None:
             if set([c for c in head.columns]) != set(column_names):
                 raise ValueError("head.columns != column_names")
@@ -600,6 +569,7 @@ class TableDescription(ViewRepresentation):
         self.limit_was = limit_was
         self.sql_meta = sql_meta
         self.table_name = table_name
+        self.nrows = nrows
         if isinstance(column_names, str):
             column_names = [column_names]
         self.column_names = [c for c in column_names]
@@ -608,8 +578,7 @@ class TableDescription(ViewRepresentation):
             self.column_types = column_types.copy()
         if qualifiers is None:
             qualifiers = {}
-        if not isinstance(qualifiers, dict):
-            raise TypeError("qualifiers must be a dictionary")
+        assert isinstance(qualifiers, dict)
         self.qualifiers = qualifiers.copy()
         self.key = ""
         if self.table_name is not None:
@@ -664,7 +633,7 @@ class TableDescription(ViewRepresentation):
         return True
 
     def to_python_implementation(self, *, indent=0, strict=True, print_sources=True):
-        spacer = ' '
+        spacer = " "
         if indent >= 0:
             spacer = "\n " + " " * indent
         column_limit = 20
@@ -706,16 +675,14 @@ class TableDescription(ViewRepresentation):
     def eval_implementation(self, *, data_map, data_model, narrow):
         if data_model is None:
             raise ValueError("Expected data_model to not be None")
-        return data_model.table_step(
-            op=self, data_map=data_map, narrow=narrow
-        )
+        return data_model.table_step(op=self, data_map=data_map, narrow=narrow)
 
     def columns_used_from_sources(self, using=None):
         return []  # no inputs to table description
 
-    def to_near_sql_implementation(self, db_model, *, using, temp_id_source):
+    def to_near_sql_implementation(self, db_model, *, using, temp_id_source, sql_format_options=None):
         return db_model.table_def_to_sql(
-            self, using=using, temp_id_source=temp_id_source
+            self, using=using, temp_id_source=temp_id_source, sql_format_options=sql_format_options
         )
 
     def __str__(self):
@@ -739,33 +706,49 @@ class TableDescription(ViewRepresentation):
         return self.key.__hash__()
 
 
-def describe_table(d, table_name="data_frame",
-                   *,
-                   qualifiers=None,
-                   sql_meta=None,
-                   column_types=None,
-                   row_limit=7):
-    # Expect a pandas.DataFrame style object
+def describe_table(
+    d,
+    table_name="data_frame",
+    *,
+    qualifiers=None,
+    sql_meta=None,
+    column_types=None,
+    row_limit=7,
+    keep_sample=True
+):
+    """
+    :param d: pandas table to describe
+    :param table_name: name of table
+    :param qualifiers: optional able qualifiers
+    :param sql_meta: optional sql meta information map
+    :param row_limit: how many rows to sampe
+    :param keep_sample: logical, if True retain head of table
+    :return: TableDescription
+    """
     assert d is not None
     assert not isinstance(d, OperatorPlatform)
     assert not isinstance(d, ViewRepresentation)
     column_names = [c for c in d.columns]
     if column_types is None:
-        if d.shape[0] > 0:
-            column_types = {k: type(d.loc[0, k]) for k in column_names}
-    if d.shape[0] > row_limit:
-        head = d.iloc[range(row_limit), :].copy()
-    else:
-        head = d.copy()
-    head.reset_index(drop=True, inplace=True)
+        column_types = data_algebra.util.guess_column_types(d)
+    head = None
+    nrows = None
+    if keep_sample:
+        nrows = d.shape[0]
+        if nrows > row_limit:
+            head = d.iloc[range(row_limit), :].copy()
+        else:
+            head = d.copy()
+        head.reset_index(drop=True, inplace=True)
     return TableDescription(
-        table_name,
-        column_names,
+        table_name=table_name,
+        column_names=column_names,
         column_types=column_types,
         qualifiers=qualifiers,
         sql_meta=sql_meta,
         head=head,
         limit_was=row_limit,
+        nrows=nrows,
     )
 
 
@@ -846,21 +829,21 @@ class ExtendNode(ViewRepresentation):
                         + "'"
                     )
                 if len(opk.args) > 1:
-                    raise ValueError(
-                        "in windowed situations only simple operators are allowed, "
-                        + "'"
-                        + k
-                        + "': '"
-                        + str(opk)
-                        + "' term is too complex an expression"
-                    )
+                    for i in range(1, len(opk.args)):
+                        if not isinstance(opk.args[i], data_algebra.expr_rep.Value):
+                            raise ValueError(
+                                "in windowed situations only simple operators are allowed, "
+                                + "'"
+                                + k
+                                + "': '"
+                                + str(opk)
+                                + "' term is too complex an expression"
+                            )
                 if len(opk.args) > 0:
                     if isinstance(opk.args[0], data_algebra.expr_rep.ColumnReference):
                         value_name = opk.args[0].column_name
                         if value_name not in source.column_set:
-                            raise ValueError(
-                                value_name + " not in source column set"
-                            )
+                            raise ValueError(value_name + " not in source column set")
                     else:
                         if not isinstance(opk.args[0], data_algebra.expr_rep.Value):
                             raise ValueError(
@@ -871,17 +854,19 @@ class ExtendNode(ViewRepresentation):
                                 + str(opk)
                                 + "' term is too complex an expression"
                             )
-                if (ordered_windowed_situation and
-                        (opk.op in data_algebra.expr_rep.fn_names_that_contradict_ordered_windowed_situation)):
+                if ordered_windowed_situation and (
+                    opk.op
+                    in data_algebra.expr_rep.fn_names_that_contradict_ordered_windowed_situation
+                ):
                     raise ValueError(
-                        str(opk)
-                        + "' is not allowed in an ordered windowed situation"
+                        str(opk) + "' is not allowed in an ordered windowed situation"
                     )
-                if ((not ordered_windowed_situation) and
-                        (opk.op in data_algebra.expr_rep.fn_names_that_imply_ordered_windowed_situation)):
+                if (not ordered_windowed_situation) and (
+                    opk.op
+                    in data_algebra.expr_rep.fn_names_that_imply_ordered_windowed_situation
+                ):
                     raise ValueError(
-                        str(opk)
-                        + "' is not allowed in not-ordered windowed situation"
+                        str(opk) + "' is not allowed in not-ordered windowed situation"
                     )
         ViewRepresentation.__init__(
             self, column_names=column_names, sources=[source], node_name="ExtendNode"
@@ -923,10 +908,16 @@ class ExtendNode(ViewRepresentation):
             for (k, opk) in self.ops.items():
                 if len(opk.args) > 0:
                     if len(opk.args) > 1:
-                        raise ValueError("window function with more than one argument")
+                        for i in range(1, len(opk.args)):
+                            if not isinstance(opk.args[i], data_algebra.expr_rep.Value):
+                                raise ValueError("window function with more than one non-value argument")
                     for i in range(len(opk.args)):
-                        if not (isinstance(opk.args[0], data_algebra.expr_rep.ColumnReference)
-                                or isinstance(opk.args[0], data_algebra.expr_rep.Value)):
+                        if not (
+                            isinstance(
+                                opk.args[0], data_algebra.expr_rep.ColumnReference
+                            )
+                            or isinstance(opk.args[0], data_algebra.expr_rep.Value)
+                        ):
                             raise ValueError(
                                 "window expression argument must be a column or value: "
                                 + str(k)
@@ -948,7 +939,7 @@ class ExtendNode(ViewRepresentation):
         return [columns_we_take]
 
     def to_python_implementation(self, *, indent=0, strict=True, print_sources=True):
-        spacer = ' '
+        spacer = " "
         if indent >= 0:
             spacer = "\n   " + " " * indent
         s = ""
@@ -976,15 +967,14 @@ class ExtendNode(ViewRepresentation):
         s = s + ")"
         return s
 
-    def to_near_sql_implementation(self, db_model, *, using, temp_id_source):
-        return db_model.extend_to_sql(self, using=using, temp_id_source=temp_id_source)
+    def to_near_sql_implementation(self, db_model, *, using, temp_id_source, sql_format_options=None):
+        return db_model.extend_to_sql(self, using=using, temp_id_source=temp_id_source,
+                                      sql_format_options=sql_format_options)
 
     def eval_implementation(self, *, data_map, data_model, narrow):
         if data_model is None:
             raise ValueError("Expected data_model to not be None")
-        return data_model.extend_step(
-            op=self, data_map=data_map, narrow=narrow
-        )
+        return data_model.extend_step(op=self, data_map=data_map, narrow=narrow)
 
 
 class ProjectNode(ViewRepresentation):
@@ -1027,27 +1017,30 @@ class ProjectNode(ViewRepresentation):
                         + str(opk)
                     )
                 if len(opk.args) > 0:
-                    if not (isinstance(opk.args[0], data_algebra.expr_rep.ColumnReference)
-                            or isinstance(opk.args[0], data_algebra.expr_rep.Value)):
+                    if not (
+                        isinstance(opk.args[0], data_algebra.expr_rep.ColumnReference)
+                        or isinstance(opk.args[0], data_algebra.expr_rep.Value)
+                    ):
                         raise ValueError(
                             "windows expression argument must be a column or value: "
                             + str(k)
                             + ": "
                             + str(opk)
                         )
-                if opk.op in data_algebra.expr_rep.fn_names_that_imply_ordered_windowed_situation:
-                    raise ValueError(
-                        str(opk)
-                        + "' is not allowed in project"
-                    )
+                if (
+                    opk.op
+                    in data_algebra.expr_rep.fn_names_that_imply_ordered_windowed_situation
+                ):
+                    raise ValueError(str(opk) + "' is not allowed in project")
+                if (
+                    opk.op
+                    in data_algebra.expr_rep.fn_names_not_allowed_in_project
+                ):
+                    raise ValueError(str(opk) + "' is not allowed in project")
             else:
-                if not isinstance(opk, data_algebra.user_fn.FnTerm):
-                    raise ValueError(
-                        "non-aggregated expression in project: "
-                        + str(k)
-                        + ": "
-                        + str(opk)
-                    )
+                raise ValueError(
+                    "non-aggregated expression in project: " + str(k) + ": " + str(opk)
+                )
             # TODO: check op is in list of aggregators
             # Note: non-aggregators making through will be caught by table shape check
 
@@ -1088,7 +1081,7 @@ class ProjectNode(ViewRepresentation):
         return [columns_we_take]
 
     def to_python_implementation(self, *, indent=0, strict=True, print_sources=True):
-        spacer = ' '
+        spacer = " "
         if indent >= 0:
             spacer = "\n   " + " " * indent
         s = ""
@@ -1115,15 +1108,14 @@ class ProjectNode(ViewRepresentation):
         s = s + ")"
         return s
 
-    def to_near_sql_implementation(self, db_model, *, using, temp_id_source):
-        return db_model.project_to_sql(self, using=using, temp_id_source=temp_id_source)
+    def to_near_sql_implementation(self, db_model, *, using, temp_id_source, sql_format_options=None):
+        return db_model.project_to_sql(self, using=using, temp_id_source=temp_id_source,
+                                       sql_format_options=sql_format_options)
 
     def eval_implementation(self, *, data_map, data_model, narrow):
         if data_model is None:
             raise ValueError("Expected data_model to not be None")
-        return data_model.project_step(
-            op=self, data_map=data_map, narrow=narrow
-        )
+        return data_model.project_step(op=self, data_map=data_map, narrow=narrow)
 
 
 class SelectRowsNode(ViewRepresentation):
@@ -1180,17 +1172,15 @@ class SelectRowsNode(ViewRepresentation):
         s = s + ("select_rows(" + self.expr.to_python().__repr__() + ")")
         return s
 
-    def to_near_sql_implementation(self, db_model, *, using, temp_id_source):
+    def to_near_sql_implementation(self, db_model, *, using, temp_id_source, sql_format_options=None):
         return db_model.select_rows_to_sql(
-            self, using=using, temp_id_source=temp_id_source
+            self, using=using, temp_id_source=temp_id_source, sql_format_options=sql_format_options
         )
 
     def eval_implementation(self, *, data_map, data_model, narrow):
         if data_model is None:
             raise ValueError("Expected data_model to not be None")
-        return data_model.select_rows_step(
-            op=self, data_map=data_map, narrow=narrow
-        )
+        return data_model.select_rows_step(op=self, data_map=data_map, narrow=narrow)
 
 
 class SelectColumnsNode(ViewRepresentation):
@@ -1251,17 +1241,15 @@ class SelectColumnsNode(ViewRepresentation):
         s = s + ("select_columns(" + self.column_selection.__repr__() + ")")
         return s
 
-    def to_near_sql_implementation(self, db_model, *, using, temp_id_source):
+    def to_near_sql_implementation(self, db_model, *, using, temp_id_source, sql_format_options=None):
         return db_model.select_columns_to_sql(
-            self, using=using, temp_id_source=temp_id_source
+            self, using=using, temp_id_source=temp_id_source, sql_format_options=sql_format_options
         )
 
     def eval_implementation(self, *, data_map, data_model, narrow):
         if data_model is None:
             raise ValueError("Expected data_model to not be None")
-        return data_model.select_columns_step(
-            op=self, data_map=data_map, narrow=narrow
-        )
+        return data_model.select_columns_step(op=self, data_map=data_map, narrow=narrow)
 
 
 class DropColumnsNode(ViewRepresentation):
@@ -1322,17 +1310,15 @@ class DropColumnsNode(ViewRepresentation):
         s = s + ("drop_columns(" + self.column_deletions.__repr__() + ")")
         return s
 
-    def to_near_sql_implementation(self, db_model, *, using, temp_id_source):
+    def to_near_sql_implementation(self, db_model, *, using, temp_id_source, sql_format_options=None):
         return db_model.drop_columns_to_sql(
-            self, using=using, temp_id_source=temp_id_source
+            self, using=using, temp_id_source=temp_id_source, sql_format_options=sql_format_options
         )
 
     def eval_implementation(self, *, data_map, data_model, narrow):
         if data_model is None:
             raise ValueError("Expected data_model to not be None")
-        return data_model.drop_columns_step(
-            op=self, data_map=data_map, narrow=narrow
-        )
+        return data_model.drop_columns_step(op=self, data_map=data_map, narrow=narrow)
 
 
 class OrderRowsNode(ViewRepresentation):
@@ -1405,15 +1391,14 @@ class OrderRowsNode(ViewRepresentation):
         s = s + ")"
         return s
 
-    def to_near_sql_implementation(self, db_model, *, using, temp_id_source):
-        return db_model.order_to_sql(self, using=using, temp_id_source=temp_id_source)
+    def to_near_sql_implementation(self, db_model, *, using, temp_id_source, sql_format_options=None):
+        return db_model.order_to_sql(self, using=using, temp_id_source=temp_id_source,
+                                     sql_format_options=sql_format_options)
 
     def eval_implementation(self, *, data_map, data_model, narrow):
         if data_model is None:
             raise ValueError("Expected data_model to not be None")
-        return data_model.order_rows_step(
-            op=self, data_map=data_map, narrow=narrow
-        )
+        return data_model.order_rows_step(op=self, data_map=data_map, narrow=narrow)
 
     # short-cut main interface
 
@@ -1500,15 +1485,14 @@ class RenameColumnsNode(ViewRepresentation):
         s = s + ("rename_columns(" + self.column_remapping.__repr__() + ")")
         return s
 
-    def to_near_sql_implementation(self, db_model, *, using, temp_id_source):
-        return db_model.rename_to_sql(self, using=using, temp_id_source=temp_id_source)
+    def to_near_sql_implementation(self, db_model, *, using, temp_id_source, sql_format_options=None):
+        return db_model.rename_to_sql(self, using=using, temp_id_source=temp_id_source,
+                                      sql_format_options=sql_format_options)
 
     def eval_implementation(self, *, data_map, data_model, narrow):
         if data_model is None:
             raise ValueError("Expected data_model to not be None")
-        return data_model.rename_columns_step(
-            op=self, data_map=data_map, narrow=narrow
-        )
+        return data_model.rename_columns_step(op=self, data_map=data_map, narrow=narrow)
 
 
 class NaturalJoinNode(ViewRepresentation):
@@ -1520,7 +1504,9 @@ class NaturalJoinNode(ViewRepresentation):
         a_tables = a.get_tables()
         b_tables = b.get_tables()
         if by is None:
-            raise ValueError("Must specify by in natural joins ([] for empty conditions)")
+            raise ValueError(
+                "Must specify by in natural joins ([] for empty conditions)"
+            )
         common_keys = set(a_tables.keys()).intersection(b_tables.keys())
         for k in common_keys:
             if not a_tables[k].same_table(b_tables[k]):
@@ -1552,7 +1538,7 @@ class NaturalJoinNode(ViewRepresentation):
         )
         self.by = by
         self.jointype = data_algebra.expr_rep.standardize_join_type(jointype)
-        if (self.jointype == 'CROSS') and (len(self.by) != 0):
+        if (self.jointype == "CROSS") and (len(self.by) != 0):
             raise ValueError("CROSS joins must have an empty 'by' list")
         self.get_tables()  # causes a throw if left and right table descriptions are inconsistent
 
@@ -1603,17 +1589,15 @@ class NaturalJoinNode(ViewRepresentation):
         )
         return s
 
-    def to_near_sql_implementation(self, db_model, *, using, temp_id_source):
+    def to_near_sql_implementation(self, db_model, *, using, temp_id_source, sql_format_options=None):
         return db_model.natural_join_to_sql(
-            self, using=using, temp_id_source=temp_id_source
+            self, using=using, temp_id_source=temp_id_source, sql_format_options=sql_format_options
         )
 
     def eval_implementation(self, *, data_map, data_model, narrow):
         if data_model is None:
             raise ValueError("Expected data_model to not be None")
-        return data_model.natural_join_step(
-            op=self, data_map=data_map, narrow=narrow
-        )
+        return data_model.natural_join_step(op=self, data_map=data_map, narrow=narrow)
 
 
 class ConcatRowsNode(ViewRepresentation):
@@ -1636,7 +1620,12 @@ class ConcatRowsNode(ViewRepresentation):
         if id_column is not None and id_column in sources[0].column_names:
             raise ValueError("id_column should not be an input table column name")
         column_names = sources[0].column_names.copy()
+        if (isinstance(a, TableDescription) and (a.column_types is not None) and (len(a.column_types) > 0)
+            and isinstance(b, TableDescription) and (b.column_types is not None) and (len(b.column_types) > 0)):
+            for c in column_names:
+                assert len({a.column_types[c]}.union({b.column_types[c]}) - {type(None)}) <= 1
         if id_column is not None:
+            assert id_column not in column_names
             column_names.append(id_column)
         ViewRepresentation.__init__(
             self, column_names=column_names, sources=sources, node_name="ConcatRowsNode"
@@ -1703,17 +1692,15 @@ class ConcatRowsNode(ViewRepresentation):
         )
         return s
 
-    def to_near_sql_implementation(self, db_model, *, using, temp_id_source):
+    def to_near_sql_implementation(self, db_model, *, using, temp_id_source, sql_format_options=None):
         return db_model.concat_rows_to_sql(
-            self, using=using, temp_id_source=temp_id_source
+            self, using=using, temp_id_source=temp_id_source, sql_format_options=sql_format_options
         )
 
     def eval_implementation(self, *, data_map, data_model, narrow):
         if data_model is None:
             raise ValueError("Expected data_model to not be None")
-        return data_model.concat_rows_step(
-            op=self, data_map=data_map, narrow=narrow
-        )
+        return data_model.concat_rows_step(op=self, data_map=data_map, narrow=narrow)
 
 
 class ConvertRecordsNode(ViewRepresentation):
@@ -1739,7 +1726,7 @@ class ConvertRecordsNode(ViewRepresentation):
         temp_id_source[0] = temp_id_source[0] + 1
         res = TableDescription(
             table_name=view_name,
-            column_names=self.record_map.blocks_out.control_table.columns
+            column_names=self.record_map.blocks_out.control_table.columns,
         )
         return res
 
@@ -1773,16 +1760,18 @@ class ConvertRecordsNode(ViewRepresentation):
         s = s + ")"
         return s
 
-    def to_near_sql_implementation(self, db_model, *, using, temp_id_source):
+    def to_near_sql_implementation(self, db_model, *, using, temp_id_source, sql_format_options=None):
         if temp_id_source is None:
             temp_id_source = [0]
         # TODO: narrow to what we are using
         sub_query = self.sources[0].to_near_sql_implementation(
-            db_model=db_model, using=None, temp_id_source=temp_id_source
+            db_model=db_model, using=None, temp_id_source=temp_id_source, sql_format_options=sql_format_options
         )
         # claims to use all columns
         query = sub_query.to_sql(
-            columns=self.columns_used_from_sources()[0], db_model=db_model, force_sql=True
+            columns=self.columns_used_from_sources()[0],
+            db_model=db_model,
+            force_sql=True,
         )
         control_out_table = None
         if self.record_map.blocks_in is not None:
@@ -1805,16 +1794,19 @@ class ConvertRecordsNode(ViewRepresentation):
         if control_out_table is not None:
             if control_out_table.key in temp_tables.keys():
                 raise ValueError(
-                    "key collision in temp_tables construction: " + control_out_table.key
+                    "key collision in temp_tables construction: "
+                    + control_out_table.key
                 )
-            temp_tables[control_out_table.key] = self.record_map.blocks_out.control_table
+            temp_tables[
+                control_out_table.key
+            ] = self.record_map.blocks_out.control_table
         near_sql = data_algebra.near_sql.NearSQLq(
             quoted_query_name=db_model.quote_identifier(view_name),
             prev_quoted_query_name=db_model.quote_identifier(prev_view_name),
             query=query,
             terms=terms,
             temp_tables=temp_tables,
-            annotation='convert records'
+            annotation="convert records",
         )
         return near_sql
 
