@@ -14,8 +14,9 @@ import data_algebra.util
 #  http://tomerfiliba.com/blog/Infix-Operators/
 
 
-# list of window/aggregation functons that must be windowed/aggegated
+# list of window/aggregation functions that must be windowed/aggregated
 # (note some other functions work in more than one mode)
+# noinspection SpellCheckingInspection
 fn_names_that_imply_windowed_situation = {
     "all",
     "any",
@@ -55,6 +56,7 @@ fn_names_that_imply_windowed_situation = {
 }
 
 
+# noinspection SpellCheckingInspection
 fn_names_that_imply_ordered_windowed_situation = {
     "cumcount",
     "cummax",
@@ -62,11 +64,21 @@ fn_names_that_imply_ordered_windowed_situation = {
     "cumprod",
     "cumsum",
     "row_number",
+    "shift",
+    "lag",
+    "lead",
 }
 
 
-# a competing idea sould be to remove ordering if
+# noinspection SpellCheckingInspection
+fn_names_not_allowed_in_project = {
+    'ngroup',
+}
+
+
+# a competing idea should be to remove ordering if
 # operator is one of these (instead of forbid)
+# noinspection SpellCheckingInspection
 fn_names_that_contradict_ordered_windowed_situation = {
     "count",
     "max",
@@ -119,7 +131,9 @@ class PreTerm(ABC):
     # emitters
 
     def to_python(self, *, want_inline_parens=False):
-        raise NotImplementedError("base class method called")  # https://docs.python.org/3/library/exceptions.html
+        raise NotImplementedError(
+            "base class method called"
+        )  # https://docs.python.org/3/library/exceptions.html
 
     def to_source(self, *, want_inline_parens=False, dialect="Python"):
         if dialect == "Python":
@@ -136,6 +150,7 @@ class PreTerm(ABC):
         return self.to_python(want_inline_parens=False)
 
 
+# noinspection SpellCheckingInspection
 class Term(PreTerm, ABC):
     """
     Abstract intermediate class with combination ability
@@ -148,34 +163,30 @@ class Term(PreTerm, ABC):
 
     def __op_expr__(self, op, other, *, inline=True, method=False):
         """binary expression"""
-        if not isinstance(op, str):
-            raise TypeError("op is supposed to be a string")
+        assert isinstance(op, str)
         if not isinstance(other, Term):
-            other = _enc_value(other)
+            other = enc_value(other)
         return Expression(op, (self, other), inline=inline, method=method)
 
     def __rop_expr__(self, op, other):
         """reversed binary expression"""
-        if not isinstance(op, str):
-            raise TypeError("op is supposed to be a string")
+        assert isinstance(op, str)
         if not isinstance(other, Term):
-            other = _enc_value(other)
+            other = enc_value(other)
         return Expression(op, (other, self), inline=True)
 
     def __uop_expr__(self, op, *, params=None, inline=False):
         """unary expression"""
-        if not isinstance(op, str):
-            raise TypeError("op is supposed to be a string")
-        return Expression(op, (self,), params=params, inline=inline)
+        assert isinstance(op, str)
+        return Expression(op, (self,), params=params, inline=inline, method=not inline)
 
     def __triop_expr__(self, op, x, y, inline=False, method=False):
         """three argument expression"""
-        if not isinstance(op, str):
-            raise TypeError("op is supposed to be a string")
+        assert isinstance(op, str)
         if not isinstance(x, Term):
-            x = _enc_value(x)
+            x = enc_value(x)
         if not isinstance(y, Term):
-            y = _enc_value(y)
+            y = enc_value(y)
         return Expression(op, (self, x, y), inline=inline, method=method)
 
     # try to get at == and other comparison operators
@@ -479,8 +490,11 @@ class Term(PreTerm, ABC):
 
     # pandas shift
 
-    def shift(self):
-        return self.__uop_expr__("shift")
+    def shift(self, periods=None):
+        if periods is None:
+            periods = Value(1)
+        assert isinstance(periods, Value)  # digits control
+        return self.__op_expr__("shift", periods, inline=False, method=True)
 
     # our ad-hoc definitions
 
@@ -497,20 +511,108 @@ class Term(PreTerm, ABC):
         return self.__op_expr__("is_in", x, inline=False, method=True)
 
     def concat(self, x):
+        # TODO: see if we can format back to infix notation
         return self.__op_expr__("concat", x, inline=False, method=True)
 
     def coalesce(self, x):
+        # TODO: see if we can format back to infix notation
         return self.__op_expr__("coalesce", x, inline=False, method=True)
 
     def co_equalizer(self, x):
         return self.__op_expr__("co_equalizer", x, inline=False, method=True)
 
+    # fns that had been in bigquery_user_fns
+
+    def as_int64(self):
+        return self.__uop_expr__("as_int64")
+
+    def as_str(self):
+        return self.__uop_expr__("as_str")
+
+    def trimstr(self, start, stop):
+        assert isinstance(start, Value)
+        assert isinstance(stop, Value)
+        return self.__triop_expr__(
+            "trimstr", x=start, y=stop, inline=False, method=True
+        )
+
+    def coalesce_0(self):
+        return self.coalesce(Value(0))
+
+    def datetime_to_date(self):
+        return self.__uop_expr__("datetime_to_date")
+
+    def parse_datetime(self, format=None):
+        if format is None:
+            format = Value("%Y-%m-%d %H:%M:%S")
+        assert isinstance(format, Value)
+        return self.__op_expr__(
+            "parse_datetime", other=format, inline=False, method=True
+        )
+
+    def parse_date(self, format=None):
+        if format is None:
+            format = Value("%Y-%m-%d")
+        return self.__op_expr__("parse_date", other=format, inline=False, method=True)
+
+    def format_datetime(self, format=None):
+        if format is None:
+            format = Value("%Y-%m-%d %H:%M:%S")
+        assert isinstance(format, Value)
+        return self.__op_expr__(
+            "format_datetime", other=format, inline=False, method=True
+        )
+
+    def format_date(self, format=None):
+        if format is None:
+            format = Value("%Y-%m-%d")
+        return self.__op_expr__("format_date", other=format, inline=False, method=True)
+
+    def dayofweek(self):
+        return self.__uop_expr__("dayofweek")
+
+    def dayofyear(self):
+        return self.__uop_expr__("dayofyear")
+
+    def dayofmonth(self):
+        return self.__uop_expr__("dayofmonth")
+
+    def weekofyear(self):
+        return self.__uop_expr__("weekofyear")
+
+    def month(self):
+        return self.__uop_expr__("month")
+
+    def quarter(self):
+        return self.__uop_expr__("quarter")
+
+    def year(self):
+        return self.__uop_expr__("year")
+
+    def timestamp_diff(self, other):
+        return self.__op_expr__("timestamp_diff", other, inline=False, method=True)
+
+    def date_diff(self, other):
+        return self.__op_expr__("date_diff", other, inline=False, method=True)
+
+    # noinspection PyPep8Naming
+    def base_Sunday(self):
+        return self.__uop_expr__("base_Sunday")
+
+
+def kop_expr(op, args, inline=False, method=False):
+    """three argument expression"""
+    assert isinstance(op, str)
+    args = [(ai if isinstance(ai, Term) else enc_value(ai)) for ai in args]
+    return Expression(op, args, inline=inline, method=method)
+
 
 class Value(Term):
     def __init__(self, value):
-        allowed = [int, float, str, bool]
-        if not any([isinstance(value, tp) for tp in allowed]):
-            raise TypeError("value type must be one of: " + str(allowed))
+        allowed = {data_algebra.util.map_type_to_canonical(t) for t in [int, float, str, bool]}
+        disaallowed = {data_algebra.util.map_type_to_canonical(type(value))} - allowed
+        if len(disaallowed) != 0:
+            raise TypeError("value type must be one of: " + str(allowed) + ', saw ' + str(list(disaallowed)[0]))
         self.value = value
         Term.__init__(self)
 
@@ -553,8 +655,7 @@ class UnQuotedStr(str):
 class ListTerm(PreTerm):
     # derived from PreTerm as this is not combinable
     def __init__(self, value):
-        if not isinstance(value, (list, tuple)):
-            raise TypeError("value type must be a list or tuple")
+        assert isinstance(value, (list, tuple))
         self.value = [v for v in value]  # copy and standardize to a list
         PreTerm.__init__(self)
 
@@ -566,11 +667,12 @@ class ListTerm(PreTerm):
 
     def get_views(self):
         views = list()
-        for ai in self.args:
-            vi = ai.get_views()
-            for vii in vi:
-                if vii not in views:  # expect list to be of size zero or one
-                    views.append(vii)
+        for ai in self.value:
+            if isinstance(ai, PreTerm):
+                vi = ai.get_views()
+                for vii in vi:
+                    if vii not in views:  # expect list to be of size zero or one
+                        views.append(vii)
         return views
 
     def replace_view(self, view):
@@ -587,7 +689,7 @@ class ListTerm(PreTerm):
         return res
 
     def to_python(self, *, want_inline_parens=False):
-        def li_to_python(value, *, want_inline_parens=False):
+        def li_to_python(value):
             try:
                 return value.to_python(want_inline_parens=want_inline_parens)
             except AttributeError:
@@ -597,7 +699,7 @@ class ListTerm(PreTerm):
             "["
             + ", ".join(
                 [
-                    li_to_python(ai, want_inline_parens=want_inline_parens)
+                    li_to_python(ai)
                     for ai in self.value
                 ]
             )
@@ -609,7 +711,7 @@ class ListTerm(PreTerm):
             ti.get_column_names(columns_seen)
 
 
-def _enc_value(value):
+def enc_value(value):
     if isinstance(value, PreTerm):
         return value
     if callable(value):
@@ -628,11 +730,14 @@ class ColumnReference(Term):
     def __init__(self, view, column_name):
         self.view = view
         self.column_name = column_name
-        if not isinstance(column_name, str):
-            raise TypeError("column_name must be a string")
+        assert isinstance(column_name, str)
         if view is not None:
             if column_name not in view.column_set:
-                raise KeyError("column_name '" + str(column_name) + "' must be a column of the given view")
+                raise KeyError(
+                    "column_name '"
+                    + str(column_name)
+                    + "' must be a column of the given view"
+                )
         Term.__init__(self)
 
     def evaluate(self, data_frame):
@@ -661,19 +766,65 @@ class ColumnReference(Term):
         columns_seen.add(self.column_name)
 
 
+# noinspection SpellCheckingInspection
+def _can_find_method_by_name(op):
+    assert isinstance(op, str)
+    # from populate_specials
+    if op in {
+        "_count",
+        "count",
+        "row_number",
+        "_row_number",
+        "_size",
+        "size",
+        "connected_components",
+        "_ngroup",
+        "ngroup",
+    }:
+        return True
+    # check user fns
+    # first check chosen mappings
+    try:
+        # noinspection PyUnusedLocal
+        check_val = data_algebra.default_data_model.user_fun_map[op]  # for KeyError
+        return True
+    except KeyError:
+        pass
+    # check chosen mappings
+    try:
+        # noinspection PyUnusedLocal
+        check_val = data_algebra.default_data_model.impl_map[op]  # for KeyError
+        return True
+    except KeyError:
+        pass
+    # now see if argument (usually Pandas) can do this
+    # doubt we hit in this, as most exposed methods are window methods
+    try:
+        method = getattr(Value(0), op)
+        if callable(method):
+            return True
+    except AttributeError:
+        pass
+    # new see if numpy can do this
+    try:
+        fn = numpy.__dict__[op]
+        if callable(fn):
+            return True
+    except KeyError:
+        pass
+    return False
+
+
 class Expression(Term):
     def __init__(self, op, args, *, params=None, inline=False, method=False):
-        if not isinstance(op, str):
-            raise TypeError("op is supposed to be a string")
+        assert isinstance(op, str)
+        if not _can_find_method_by_name(op):
+            raise KeyError(f"can't find implementation for function/method {op}")
         if inline:
             if method:
                 raise ValueError("can't set both inline and method")
-            if len(args) > 2:
-                raise ValueError(
-                    "must have no more than two arguments if inline is True"
-                )
         self.op = op
-        self.args = [_enc_value(ai) for ai in args]
+        self.args = [enc_value(ai) for ai in args]
         self.params = params
         self.inline = inline
         self.method = method
@@ -715,7 +866,11 @@ class Expression(Term):
     def replace_view(self, view):
         new_args = [oi.replace_view(view) for oi in self.args]
         return Expression(
-            op=self.op, args=new_args, params=self.params, inline=self.inline
+            op=self.op,
+            args=new_args,
+            params=self.params,
+            inline=self.inline,
+            method=self.method,
         )
 
     def get_column_names(self, columns_seen):
@@ -724,9 +879,17 @@ class Expression(Term):
 
     def evaluate(self, data_frame):
         args = [ai.evaluate(data_frame) for ai in self.args]
+        # check user fns
         # first check chosen mappings
         try:
-            return data_algebra.default_data_model.impl_map[self.op](*args)
+            method_to_call = data_algebra.default_data_model.user_fun_map[self.op]
+            return method_to_call(*args)
+        except KeyError:
+            pass
+        # check chosen mappings
+        try:
+            method_to_call = data_algebra.default_data_model.impl_map[self.op]
+            return method_to_call(*args)
         except KeyError:
             pass
         # now see if argument (usually Pandas) can do this
@@ -734,7 +897,7 @@ class Expression(Term):
         try:
             method = getattr(args[0], self.op)
             if callable(method):
-                return method(*args[1: ])
+                return method(*args[1:])
         except AttributeError:
             pass
         # new see if numpy can do this
@@ -744,28 +907,33 @@ class Expression(Term):
                 return fn(*args)
         except KeyError:
             pass
+        raise KeyError(f"function {self.op} not found")
 
     def to_python(self, *, want_inline_parens=False):
         subs = [ai.to_python(want_inline_parens=True) for ai in self.args]
         if len(subs) <= 0:
-            return "_" + self.op + "()"
+            return self.op + "()"
         if len(subs) == 1:
             if self.inline:
                 return self.op + self.args[0].to_python(want_inline_parens=True)
-            if isinstance(self.args[0], ColumnReference):
-                return subs[0] + "." + self.op + "()"
-            else:
-                return "(" + subs[0] + ")." + self.op + "()"
-        if len(subs) == 2 and self.inline:
+            if self.method:
+                if isinstance(self.args[0], ColumnReference):
+                    return subs[0] + "." + self.op + "()"
+                else:
+                    return "(" + subs[0] + ")." + self.op + "()"
+        if self.inline:
+            result = ''
             if want_inline_parens:
-                return "(" + subs[0] + " " + self.op + " " + subs[1] + ")"
-            else:
-                return subs[0] + " " + self.op + " " + subs[1]
+                result = result + '('
+            result = result + (' ' + self.op + ' ').join(subs)
+            if want_inline_parens:
+                result = result + ')'
+            return result
         if self.method:
             if isinstance(self.args[0], ColumnReference):
                 return subs[0] + "." + self.op + "(" + ", ".join(subs[1:]) + ")"
             else:
-                return subs[0] + "." + self.op + "(" + ", ".join(subs[1:]) + ")"
+                return "(" + subs[0] + ")." + self.op + "(" + ", ".join(subs[1:]) + ")"
         return self.op + "(" + ", ".join(subs) + ")"
 
 
@@ -774,65 +942,27 @@ def connected_components(f, g):
     return data_algebra.expr_rep.Expression(op="connected_components", args=[f, g])
 
 
-# TODO: get rid of user_values
-def populate_specials(*, column_defs, destination, user_values=None):
-    """populate a dictionary with special values
-       column_defs is a dictionary,
-         usually formed from a ViewRepresentation.column_map
-       destination is a dictionary,
-         usually formed from a ViewRepresentation.column_map.copy()
-    """
-
-    if not isinstance(column_defs, dict):
-        raise TypeError("column_defs should be a dictionary")
-    if not isinstance(destination, dict):
-        raise TypeError("destination should be a dictionary")
-    if user_values is None:
-        user_values = {}
-    if not isinstance(user_values, dict):
-        raise TypeError("user_values should be a dictionary")
-    nd = column_defs.copy()
-    # more doc!
-    # makes these symbols available for parsing step
-    # need to enter user functions here or as methods on variables
-    # Note: a lot of the re-emitters add back the underbar, allow the non
-    # underbar names we see here.  See expr_rep.py Expression to_python.
-    destination["_row_number"] = lambda: data_algebra.expr_rep.Expression(
-        op="row_number", args=[]
-    )
-    destination["_ngroup"] = lambda: data_algebra.expr_rep.Expression(
-        op="ngroup", args=[]
-    )
-    destination["_size"] = lambda: data_algebra.expr_rep.Expression(op="size", args=[])
-    destination["connected_components"] = connected_components
-
-
 def standardize_join_type(join_str):
-    if not isinstance(join_str, str):
-        raise TypeError("Expected join_str to be a string")
+    assert isinstance(join_str, str)
     join_str = join_str.upper()
-    allowed = set(['INNER', 'LEFT', 'RIGHT', 'OUTER', 'FULL', 'CROSS'])
-    if not join_str in allowed:  # TOOO put this back in!
+    allowed = {"INNER", "LEFT", "RIGHT", "OUTER", "FULL", "CROSS"}
+    if join_str not in allowed:
         raise KeyError(f"join type {join_str} not supported")
     return join_str
 
 
+# noinspection SpellCheckingInspection
 def get_columns_used(parsed_exprs):
-    if not isinstance(parsed_exprs, dict):
-        raise TypeError(
-            "expected parsed_exprs to be a dictionary of data_algebra.expr_rep.Term(s)"
-        )
+    assert isinstance(parsed_exprs, dict)
     columns_seen = set()
     for node in parsed_exprs.values():
         node.get_column_names(columns_seen)
     return columns_seen
 
 
+# noinspection SpellCheckingInspection
 def implies_windowed(parsed_exprs):
-    if not isinstance(parsed_exprs, dict):
-        raise TypeError(
-            "expected parsed_exprs to be a dictionary of data_algebra.expr_rep.Term(s)"
-        )
+    assert isinstance(parsed_exprs, dict)
     for opk in parsed_exprs.values():  # look for aggregation functions
         if isinstance(opk, data_algebra.expr_rep.Expression):
             if opk.op in data_algebra.expr_rep.fn_names_that_imply_windowed_situation:
