@@ -1,4 +1,3 @@
-
 from abc import ABC
 
 import data_algebra.OrderedSet
@@ -14,7 +13,9 @@ class NearSQL(ABC):
     Represent SQL queries in a mostly string-form
     """
 
-    def __init__(self, *, terms, quoted_query_name, temp_tables, is_table=False, annotation=None):
+    def __init__(
+        self, *, terms, quoted_query_name, temp_tables, is_table=False, annotation=None
+    ):
         assert isinstance(terms, (dict, type(None)))
         assert isinstance(quoted_query_name, str)
         assert isinstance(temp_tables, dict)
@@ -22,7 +23,9 @@ class NearSQL(ABC):
         assert isinstance(annotation, (str, type(None)))
         self.terms = None
         if terms is not None:
-            self.terms = terms.copy()
+            assert isinstance(terms, dict)
+            if len(terms) > 0:
+                self.terms = terms.copy()
         self.quoted_query_name = quoted_query_name
         self.is_table = is_table
         self.temp_tables = temp_tables.copy()
@@ -30,12 +33,12 @@ class NearSQL(ABC):
 
     def to_bound_near_sql(self, *, columns=None, force_sql=False, constants=None):
         return NearSQLContainer(
-            near_sql=self,
-            columns=columns,
-            force_sql=force_sql,
-            constants=constants)
+            near_sql=self, columns=columns, force_sql=force_sql, constants=constants
+        )
 
-    def to_sql(self, *, columns=None, force_sql=False, constants=None, db_model, annotate=False):
+    def to_sql(
+        self, *, columns=None, force_sql=False, constants=None, db_model, sql_format_options=None
+    ):
         raise NotImplementedError("base method called")
 
     # return a list where last element is a NearSQL previous elements are (name, NearSQLContainer) pairs
@@ -50,10 +53,11 @@ class NearSQLContainer:
     NearSQL with bound in columns, force_sql, and constants decisions
     """
 
-    def __init__(self, *,
-                 near_sql, columns=None, force_sql=False, constants=None):
+    def __init__(self, *, near_sql, columns=None, force_sql=False, constants=None):
         assert isinstance(near_sql, NearSQL)
-        assert isinstance(columns, (set, data_algebra.OrderedSet.OrderedSet, list, type(None)))
+        assert isinstance(
+            columns, (set, data_algebra.OrderedSet.OrderedSet, list, type(None))
+        )
         assert isinstance(force_sql, bool)
         assert isinstance(constants, (dict, type(None)))
         self.near_sql = near_sql
@@ -66,19 +70,22 @@ class NearSQLContainer:
         if constants is not None:
             self.constants = constants.copy()
 
-    def to_sql(self, db_model, annotate=False):
+    def to_sql(self, db_model, sql_format_options=None):
         return self.near_sql.to_sql(
             columns=self.columns,
             force_sql=self.force_sql,
             constants=self.constants,
             db_model=db_model,
-            annotate=annotate)
+            sql_format_options=sql_format_options,
+        )
 
     # assemble sub-sql
-    def convert_subsql(self, *, db_model, annotate=False):
+    def convert_subsql(self, *, db_model, sql_format_options=None):
         assert isinstance(self, NearSQLContainer)
         assert isinstance(self.near_sql, NearSQL)
-        return db_model.convert_nearsql_container_subsql_(nearsql_container=self, annotate=annotate)
+        return db_model.convert_nearsql_container_subsql_(
+            nearsql_container=self, sql_format_options=sql_format_options
+        )
 
     # sequence: a list where last element is a NearSQLContainer previous elements are (name, NearSQLContainer) pairs
     # stub the replacement common table expression in a NearSQLContainer
@@ -92,15 +99,20 @@ class NearSQLContainer:
             endi = len(sequence) - 1
             last_step = sequence[endi]
             sequence[endi] = NearSQLContainer(
-                near_sql=last_step, columns=self.columns, force_sql=self.force_sql, constants=self.constants)
+                near_sql=last_step,
+                columns=self.columns,
+                force_sql=self.force_sql,
+                constants=self.constants,
+            )
             endi = len(sequence) - 1
             last_step = sequence[endi]
             stub = last_step
             if not stub.near_sql.is_table:
                 stub = NearSQLContainer(
                     near_sql=NearSQLCommonTableExpression(
-                        quoted_query_name=last_step.near_sql.quoted_query_name),
-                        force_sql=self.force_sql
+                        quoted_query_name=last_step.near_sql.quoted_query_name
+                    ),
+                    force_sql=self.force_sql,
                 )
             sequence[endi] = (last_step.near_sql.quoted_query_name, last_step)
         return stub, sequence
@@ -109,33 +121,45 @@ class NearSQLContainer:
 class NearSQLCommonTableExpression(NearSQL):
     def __init__(self, *, quoted_query_name):
         NearSQL.__init__(
-            self, terms=None, quoted_query_name=quoted_query_name, temp_tables=dict(), is_table=True)
+            self,
+            terms=None,
+            quoted_query_name=quoted_query_name,
+            temp_tables=dict(),
+            is_table=True,
+        )
 
-    def to_sql(self, *, columns=None, force_sql=False, constants=None, db_model, annotate=False):
+    def to_sql(
+        self, *, columns=None, force_sql=False, constants=None, db_model, sql_format_options=None
+    ):
         return db_model.nearsqlcte_to_sql_(
             near_sql=self,
             columns=columns,
             force_sql=force_sql,
             constants=constants,
-            annotate=annotate
+            sql_format_options=sql_format_options,
         )
 
 
 class NearSQLTable(NearSQL):
     def __init__(self, *, terms, quoted_query_name, quoted_table_name):
-        assert isinstance(terms, dict)
         NearSQL.__init__(
-            self, terms=terms, quoted_query_name=quoted_query_name, temp_tables=dict(), is_table=True
+            self,
+            terms=terms,
+            quoted_query_name=quoted_query_name,
+            temp_tables=dict(),
+            is_table=True,
         )
         self.quoted_table_name = quoted_table_name
 
-    def to_sql(self, *, columns=None, force_sql=False, constants=None, db_model, annotate=False):
+    def to_sql(
+        self, *, columns=None, force_sql=False, constants=None, db_model, sql_format_options=None
+    ):
         return db_model.nearsqltable_to_sql_(
             near_sql=self,
             columns=columns,
             force_sql=force_sql,
             constants=constants,
-            annotate=annotate
+            sql_format_options=sql_format_options,
         )
 
 
@@ -146,30 +170,43 @@ class NearSQLUnaryStep(NearSQL):
         terms,
         quoted_query_name,
         sub_sql,
-        suffix="",
+        suffix=None,
         temp_tables,
-        annotation=None
+        annotation=None,
+        mergeable=False,
+        declared_term_dependencies=None,
     ):
-        # assert isinstance(terms, dict)
+        assert isinstance(mergeable, bool)
+        assert isinstance(sub_sql, NearSQLContainer)
+        assert isinstance(suffix, (list, type(None)))
+        if not suffix is None:
+            assert all([isinstance(v, str) for v in suffix])
         NearSQL.__init__(
             self,
             terms=terms,
             quoted_query_name=quoted_query_name,
             temp_tables=temp_tables,
-            annotation=annotation
+            annotation=annotation,
         )
-        assert isinstance(sub_sql, NearSQLContainer)
-        assert isinstance(suffix,  (str, type(None)))
         self.sub_sql = sub_sql
         self.suffix = suffix
+        self.mergeable = mergeable
+        self.declared_term_dependencies = None
+        if declared_term_dependencies is not None:
+            assert isinstance(declared_term_dependencies, dict)
+            self.declared_term_dependencies = declared_term_dependencies.copy()
+        else:
+            self.mergeable = False
 
-    def to_sql(self, *, columns=None, force_sql=False, constants=None, db_model, annotate=False):
+    def to_sql(
+        self, *, columns=None, force_sql=False, constants=None, db_model, sql_format_options=None
+    ):
         return db_model.nearsqlunary_to_sql_(
             near_sql=self,
             columns=columns,
-            force_sql=force_sql,
             constants=constants,
-            annotate=annotate)
+            sql_format_options=sql_format_options,
+        )
 
     def to_with_form(self):
         if self.sub_sql.near_sql.is_table:
@@ -185,7 +222,8 @@ class NearSQLUnaryStep(NearSQL):
             sub_sql=stub,
             suffix=self.suffix,
             temp_tables=self.temp_tables,
-            annotation=self.annotation)
+            annotation=self.annotation,
+        )
         sequence.append(stubbed_step)
         return sequence
 
@@ -199,35 +237,38 @@ class NearSQLBinaryStep(NearSQL):
         sub_sql1,
         joiner,
         sub_sql2,
-        suffix="",
+        suffix=None,
         temp_tables,
         annotation=None
     ):
-        assert isinstance(terms, dict)
+        assert isinstance(sub_sql1, NearSQLContainer)
+        assert isinstance(sub_sql2, NearSQLContainer)
+        assert isinstance(suffix, (list, type(None)))
+        if not suffix is None:
+            assert all([isinstance(v, str) for v in suffix])
+        assert isinstance(joiner, str)
         NearSQL.__init__(
             self,
             terms=terms,
             quoted_query_name=quoted_query_name,
             temp_tables=temp_tables,
-            annotation=annotation
+            annotation=annotation,
         )
-        assert isinstance(sub_sql1,  NearSQLContainer)
-        assert isinstance(sub_sql2, NearSQLContainer)
-        assert isinstance(suffix,  (str, type(None)))
-        assert isinstance(joiner, str)
         self.sub_sql1 = sub_sql1
         self.joiner = joiner
         self.sub_sql2 = sub_sql2
         self.suffix = suffix
 
-    def to_sql(self, *, columns=None, force_sql=False, constants=None, db_model, annotate=False):
+    def to_sql(
+        self, *, columns=None, force_sql=False, constants=None, db_model, sql_format_options=None
+    ):
         return db_model.nearsqlbinary_to_sql_(
             near_sql=self,
             columns=columns,
             force_sql=force_sql,
             constants=constants,
-            annotate=annotate,
-            quoted_query_name=self.quoted_query_name
+            sql_format_options=sql_format_options,
+            quoted_query_name=self.quoted_query_name,
         )
 
     def to_with_form(self):
@@ -259,7 +300,8 @@ class NearSQLBinaryStep(NearSQL):
             sub_sql2=stub2,
             suffix=self.suffix,
             temp_tables=self.temp_tables,
-            annotation=self.annotation)
+            annotation=self.annotation,
+        )
         sequence.append(stubbed_step)
         return sequence
 
@@ -269,10 +311,19 @@ class NearSQLq(NearSQL):
     Adapter to wrap a pre-existing query as a NearSQL
 
     """
+
     def __init__(
-        self, *, quoted_query_name, query, terms, prev_quoted_query_name, temp_tables, annotation=None
+        self,
+        *,
+        quoted_query_name,
+        query,
+        terms,
+        prev_quoted_query_name,
+        temp_tables,
+        annotation=None
     ):
-        assert isinstance(terms, dict)
+        assert isinstance(query, str)
+        assert isinstance(prev_quoted_query_name, str)
         NearSQL.__init__(
             self,
             terms=terms,
@@ -280,16 +331,15 @@ class NearSQLq(NearSQL):
             temp_tables=temp_tables,
             annotation=annotation,
         )
-        assert isinstance(query, str)
-        assert isinstance(prev_quoted_query_name, str)
         self.query = query
         self.prev_quoted_query_name = prev_quoted_query_name
 
-    def to_sql(self, *, columns=None, force_sql=False, constants=None, db_model, annotate=False):
+    def to_sql(
+        self, *, columns=None, force_sql=False, constants=None, db_model, sql_format_options=None
+    ):
         return db_model.nearsqlq_to_sql_(
             near_sql=self,
             columns=columns,
-            force_sql=force_sql,
             constants=constants,
-            annotate=annotate,
+            sql_format_options=sql_format_options,
         )
