@@ -4,7 +4,6 @@ import numbers
 
 import numpy
 
-import data_algebra.user_fn
 import data_algebra.util
 import data_algebra.data_model
 import data_algebra.expr_rep
@@ -20,8 +19,7 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
         data_algebra.data_model.DataModel.__init__(
             self, presentation_model_name=presentation_model_name
         )
-        if not isinstance(pd, types.ModuleType):
-            raise TypeError("Expected pd to be a module")
+        assert isinstance(pd, types.ModuleType)
         self.pd = pd
 
     # utils
@@ -71,14 +69,19 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
             raise ValueError("data_map[" + op.table_name + "] was not the right type")
         # check all columns we expect are present
         columns_using = op.column_names
+        if not narrow:
+            columns_using = [c for c in df.columns]
         missing = set(columns_using) - set([c for c in df.columns])
         if len(missing) > 0:
             raise ValueError("missing required columns: " + str(missing))
         # make an index-free copy of the data to isolate side-effects and not deal with indices
-        if not narrow:
-            columns_using = [c for c in df.columns]
         res = df.loc[:, columns_using]
         res = res.reset_index(drop=True)
+        # check type compatibility
+        if (op.column_types is not None) and (len(op.column_types) > 0):
+            types_seen = data_algebra.util.guess_column_types(res)
+            for c in columns_using:
+                assert data_algebra.util.compatible_types({types_seen[c]}.union({op.column_types[c]}))
         return res
 
     def extend_step(self, op, *, data_map, narrow):
@@ -112,8 +115,7 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
 
             for (k, opk) in op.ops.items():
                 # assumes all args are column names or values, enforce this earlier
-                if (len(opk.args) > 0):
-                    assert len(opk.args) == 1
+                if len(opk.args) > 0:
                     if isinstance(opk.args[0], data_algebra.expr_rep.ColumnReference):
                         value_name = opk.args[0].column_name
                         if value_name not in col_set:
@@ -121,8 +123,10 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
                             col_set.add(value_name)
                     elif isinstance(opk.args[0], data_algebra.expr_rep.Value):
                         key = str(opk.args[0].value)
-                        if not key in data_algebra_temp_cols.keys():
-                            value_name = 'data_algebra_extend_temp_col_' + str(len(data_algebra_temp_cols))
+                        if key not in data_algebra_temp_cols.keys():
+                            value_name = "data_algebra_extend_temp_col_" + str(
+                                len(data_algebra_temp_cols)
+                            )
                             data_algebra_temp_cols[key] = value_name
                             col_list.append(value_name)
                             res[value_name] = opk.args[0].value
@@ -146,31 +150,35 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
             for (k, opk) in op.ops.items():
                 # work on a slice of the data frame
                 # TODO: document exactly which of these are available
+                # TODO: check if need any of these in project
                 if len(opk.args) <= 0:
-                    if opk.op in ["row_number", "count"]:
+                    if opk.op in {"row_number", "_row_number", "count", "_count"}:
                         subframe[k] = opframe.cumcount() + 1
-                    elif opk.op == "ngroup":
+                    elif opk.op in {"ngroup", "_ngroup"}:
                         subframe[k] = opframe.ngroup()
-                    elif opk.op == "size":
+                    elif opk.op in {"size", "_size"}:
                         subframe[k] = opframe[standin_name].transform(
                             opk.op
                         )  # Pandas transform, not data_algebra
                     else:
                         raise KeyError("not implemented: " + str(k) + ": " + str(opk))
                 else:
-                    # len(opk.args) == 1
-                    assert len(opk.args) == 1
+                    transform_args = []
+                    if len(opk.args) > 1:
+                        for i in range(1, len(opk.args)):
+                            assert isinstance(opk.args[i], data_algebra.expr_rep.Value)
+                        transform_args = [opk.args[i].value for i in range(1, len(opk.args))]
                     if isinstance(opk.args[0], data_algebra.expr_rep.ColumnReference):
                         value_name = opk.args[0].column_name
                         if value_name not in set(col_list):
                             col_list = col_list + [value_name]
                         subframe[k] = opframe[value_name].transform(
-                            opk.op
+                            opk.op, *transform_args
                         )  # Pandas transform, not data_algegra
                     elif isinstance(opk.args[0], data_algebra.expr_rep.Value):
                         value_name = data_algebra_temp_cols[str(opk.args[0].value)]
                         subframe[k] = opframe[value_name].transform(
-                            opk.op
+                            opk.op, *transform_args
                         )  # Pandas transform, not data_algegra
                     else:
                         raise ValueError("opk must be a ColumnReference or Value")
@@ -192,6 +200,16 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
         :return:
         """
         # noinspection PyUnresolvedReferences
+        assert isinstance(cols, dict)
+        # check columns all have the same length
+        if len(cols) > 1:
+            len_0 = None
+            for v in cols.items():
+                len_i = len(v)
+                if len_0 is None:
+                    len_0 = len_i
+                else:
+                    assert len_i == len_0
         return self.pd.DataFrame(cols)
 
     def project_step(self, op, *, data_map, narrow):
@@ -216,8 +234,10 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
                     pass
                 elif isinstance(opk.args[0], data_algebra.expr_rep.Value):
                     key = str(opk.args[0].value)
-                    if not key in data_algebra_temp_cols.keys():
-                        value_name = 'data_algebra_project_temp_col_' + str(len(data_algebra_temp_cols))
+                    if key not in data_algebra_temp_cols.keys():
+                        value_name = "data_algebra_project_temp_col_" + str(
+                            len(data_algebra_temp_cols)
+                        )
                         data_algebra_temp_cols[key] = value_name
                         res[value_name] = opk.args[0].value
                 else:
@@ -238,14 +258,10 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
                     value_name = str(opk.args[0])
                     if isinstance(opk.args[0], data_algebra.expr_rep.Value):
                         value_name = data_algebra_temp_cols[value_name]
-                if isinstance(opk, data_algebra.user_fn.FnTerm):
-                    fn = opk.pandas_fn
-                    vk = res[value_name].agg(fn)
+                if len(opk.args) > 0:
+                    vk = res[value_name].agg(opk.op)
                 else:
-                    if len(opk.args) > 0:
-                        vk = res[value_name].agg(opk.op)
-                    else:
-                        vk = res["_data_table_temp_col"].agg(opk.op)
+                    vk = res["_data_table_temp_col"].agg(opk.op)
                 cols[k] = vk
         else:
             cols = {"_data_table_temp_col": res["_data_table_temp_col"].agg("sum")}
@@ -260,9 +276,10 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
             return v
 
         cols = {k: promote_scalar(v) for (k, v) in cols.items()}
-        res = self.columns_to_frame(cols).reset_index(
-            drop=len(op.group_by) < 1
-        )  # grouping variables in the index
+        res = self.columns_to_frame(cols)
+        res = res.reset_index(drop=len(op.group_by) < 1)  # grouping variables in the index
+        missing_group_cols = set(op.group_by) - set(res.columns)
+        assert len(missing_group_cols) <= 0
         if "_data_table_temp_col" in res.columns:
             res = res.drop("_data_table_temp_col", 1)
         # double check shape is what we expect
@@ -331,12 +348,11 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
 
     # noinspection PyMethodMayBeStatic
     def standardize_join_code(self, jointype):
-        if not isinstance(jointype, str):
-            raise TypeError("expected jointype to be a string")
+        assert isinstance(jointype, str)
         jointype = jointype.lower()
         mp = {
             "full": "outer",
-            "cross": "outer"  # cross new to Pandas 1.2.0 December 2020
+            "cross": "outer",  # cross new to Pandas 1.2.0 December 2020
         }
         try:
             return mp[jointype]
@@ -358,12 +374,15 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
         common_cols = set([c for c in left.columns]).intersection(
             [c for c in right.columns]
         )
+        type_checks = data_algebra.util.check_columns_appear_compatible(left, right, columns=common_cols)
+        if type_checks is not None:
+            raise ValueError(f"join: incompatible column types: {type_checks}")
         by = op.by
         if by is None:
             by = []
         scratch_col = None  # extra column to prevent empty-by issues
         if len(by) <= 0:
-            scratch_col = 'data_algebra_temp_merge_col'
+            scratch_col = "data_algebra_temp_merge_col"
             by = [scratch_col]
             left[scratch_col] = 1
             right[scratch_col] = 1
@@ -401,6 +420,9 @@ class PandasModelBase(data_algebra.data_model.DataModel, ABC):
         if op.id_column is not None:
             left[op.id_column] = op.a_name
             right[op.id_column] = op.b_name
+        type_checks = data_algebra.util.check_columns_appear_compatible(left, right)
+        if type_checks is not None:
+            raise ValueError(f"concat: incompatible column types: {type_checks}")
         # noinspection PyUnresolvedReferences
         res = self.pd.concat([left, right], axis=0, ignore_index=True)
         res = res.reset_index(drop=True)
