@@ -1,10 +1,8 @@
-
 import numpy
 
 # noinspection PyUnresolvedReferences
 import data_algebra
 
-import sqlite3
 import pickle
 
 # noinspection PyUnresolvedReferences
@@ -13,10 +11,17 @@ import data_algebra.BigQuery
 import data_algebra.PostgreSQL
 import data_algebra.MySQL
 import data_algebra.SparkSQL
-import data_algebra.SparkSQL
+
 from data_algebra.data_ops import *
 
 
+# controls
+test_PostgreSQL = False  # causes an external dependency
+test_BigQuery = False  # causes an external dependency
+test_MySQL = False  # causes an external dependency
+test_Spark = False  # causes an external dependency
+
+
 def formats_to_self(ops):
     """
     Check a operator dag formats and parses back to itself.
@@ -29,10 +34,12 @@ def formats_to_self(ops):
     ops2 = eval(
         str1,
         globals(),
-        {'pd': data_algebra.default_data_model.pd}  # make our definition of pandas available
+        {
+            "pd": data_algebra.default_data_model.pd
+        },  # make our definition of pandas available
     )
     str2 = repr(ops2)
-    strings_match = str1 == str2   # probably too strict
+    strings_match = str1 == str2  # probably too strict
     ops_match = ops == ops2
     if strings_match and (not ops_match):
         raise Exception("strings match, but ops did not")
@@ -57,10 +64,8 @@ def equivalent_frames(
     # leave in extra checks as this is usually used by test code
     if local_data_model is None:
         local_data_model = data_algebra.default_data_model
-    if not isinstance(a, local_data_model.pd.DataFrame):
-        raise TypeError("Expect a to be local_data_model.pd.DataFrame")
-    if not isinstance(b, local_data_model.pd.DataFrame):
-        raise TypeError("Expect b to be local_data_model.pd.DataFrame")
+    assert isinstance(a, local_data_model.pd.DataFrame)
+    assert isinstance(b, local_data_model.pd.DataFrame)
     if a.shape != b.shape:
         return False
     if a.shape[1] < 1:
@@ -72,18 +77,18 @@ def equivalent_frames(
         a = a.reset_index(drop=True)
         b.columns = [c.lower() for c in b.columns]
         b = b.reset_index(drop=True)
-    acols = [c for c in a.columns]
-    bcols = [c for c in b.columns]
-    if set(acols) != set(bcols):
+    a_columns = [c for c in a.columns]
+    b_columns = [c for c in b.columns]
+    if set(a_columns) != set(b_columns):
         return False
     if check_column_order:
         if not all([a.columns[i] == b.columns[i] for i in range(a.shape[0])]):
             return False
     else:
         # re-order b into a's column order
-        b = b[acols]
+        b = b[a_columns]
         b = b.reset_index(drop=True)
-    for c in acols:
+    for c in a_columns:
         if local_data_model.can_convert_col_to_numeric(
             a[c]
         ) != local_data_model.can_convert_col_to_numeric(b[c]):
@@ -91,16 +96,18 @@ def equivalent_frames(
     if a.shape[0] < 1:
         return True
     if not check_row_order:
-        a = a.sort_values(by=acols)
+        a = a.sort_values(by=a_columns)
         a = a.reset_index(drop=True)
-        b = b.sort_values(by=acols)
+        b = b.sort_values(by=a_columns)
         b = b.reset_index(drop=True)
-    for c in acols:
+    for c in a_columns:
         ca = a[c]
         cb = b[c]
         if (ca is None) != (cb is None):
             return False
         if ca is not None:
+            if len(ca) != len(cb):
+                return False
             ca_null = ca.isnull()
             cb_null = cb.isnull()
             if (ca_null is None) != (cb_null is None):
@@ -113,7 +120,7 @@ def equivalent_frames(
                 if local_data_model.can_convert_col_to_numeric(a[c]):
                     ca = numpy.asarray(ca, dtype=float)
                     cb = numpy.asarray(cb, dtype=float)
-                    dif = abs(ca - cb)
+                    dif = numpy.abs(ca - cb) / numpy.maximum(numpy.maximum(ca, cb), 1)
                     if numpy.max(dif) > float_tol:
                         return False
                 else:
@@ -134,7 +141,6 @@ def check_transform_on_handles(
     check_row_order=False,
     check_parse=True,
     local_data_model=None,
-    allow_pretty=False,
 ):
     """
     Test an operator dag produces the expected result, and parses correctly.
@@ -150,7 +156,6 @@ def check_transform_on_handles(
     :param check_row_order: passed to equivalent_frames()
     :param check_parse: if True check expression parses/formats to self
     :param local_data_model: optional alternate evaluation model
-    :param allow_pretty: if True try pretty printing SQL
     :return: None, assert if there is an issue
     """
 
@@ -163,8 +168,7 @@ def check_transform_on_handles(
     assert isinstance(data, dict)
     if local_data_model is None:
         local_data_model = data_algebra.default_data_model
-    if not isinstance(ops, ViewRepresentation):
-        raise TypeError("expected ops to be a data_algebra.data_ops.ViewRepresentation")
+    assert isinstance(ops, ViewRepresentation)
     if not local_data_model.is_appropriate_data_instance(expect):
         raise TypeError("expected expect to be a local_data_model.pd.DataFrame")
     cols_used = ops.columns_used()
@@ -195,12 +199,12 @@ def check_transform_on_handles(
     if len(data) == 1:
         res_t = ops.transform(list(data.values())[0])
         if not equivalent_frames(
-                res_t,
-                expect,
-                float_tol=float_tol,
-                check_column_order=check_column_order,
-                cols_case_sensitive=cols_case_sensitive,
-                check_row_order=check_row_order,
+            res_t,
+            expect,
+            float_tol=float_tol,
+            check_column_order=check_column_order,
+            cols_case_sensitive=cols_case_sensitive,
+            check_row_order=check_row_order,
         ):
             raise ValueError("Pandas transform result did not match expect")
     # try any db paths
@@ -209,23 +213,22 @@ def check_transform_on_handles(
             to_del = set()
             temp_tables = None
             sql_statements = []
-            if allow_pretty:
-                pretty_levels = [True, False]
-            else:
-                pretty_levels = [False]
-            for annotate in [True, False]:
-                for pretty in pretty_levels:
-                    for use_with in [True, False]:
-                        temp_tables = dict()
-                        sql = db_handle.to_sql(
-                            ops,
-                            pretty=pretty,
-                            annotate=annotate,
-                            use_with=use_with,
-                            temp_tables=temp_tables)
-                        assert isinstance(sql, str)
-                        sql_statements.append(sql)
-                        # print(sql)
+            for initial_commas in [True, False]:
+                for use_with in [True, False]:
+                    sql_format_options = data_algebra.db_model.SQLFormatOptions(
+                        use_with=use_with,
+                        annotate=True,
+                        sql_indent=' ',
+                        initial_commas=initial_commas)
+                    temp_tables = dict()
+                    sql = db_handle.to_sql(
+                        ops,
+                        temp_tables=temp_tables,
+                        sql_format_options=sql_format_options,
+                    )
+                    assert isinstance(sql, str)
+                    sql_statements.append(sql)
+                    # print(sql)
             if db_handle.conn is not None:
                 for (k, v) in data.items():
                     db_handle.insert_table(v, table_name=k, allow_overwrite=True)
@@ -281,7 +284,6 @@ def check_transform(
     cols_case_sensitive=False,
     check_row_order=False,
     check_parse=True,
-    allow_pretty=True,
     models_to_skip=None,
 ):
     """
@@ -296,7 +298,6 @@ def check_transform(
     :param cols_case_sensitive: passed to equivalent_frames()
     :param check_row_order: passed to equivalent_frames()
     :param check_parse: if True check expression parses/formats to self
-    :param allow_pretty: if True try pretty printing SQL
     :param models_to_skip: None or set of model names to skip testing
     :return: nothing
     """
@@ -315,13 +316,7 @@ def check_transform(
         data_algebra.SparkSQL.SparkSQLModel().db_handle(None),
         data_algebra.MySQL.MySQLModel().db_handle(None),
         data_algebra.SQLite.example_handle(),  # actual database instance, not empty
-        ]
-
-    # controls
-    test_PostgreSQL = False  # causes an external dependency
-    test_BigQuery = False  # causes an external dependency
-    test_MySQL = False  # causes an external dependency
-    test_Spark = False  # causes an external dependency
+    ]
 
     if test_PostgreSQL:
         db_handles.append(data_algebra.PostgreSQL.example_handle())
@@ -345,7 +340,6 @@ def check_transform(
         check_row_order=check_row_order,
         check_parse=check_parse,
         db_handles=db_handles,
-        allow_pretty=allow_pretty,
     )
 
     for handle in db_handles:
